/*
 * routes.c
 * This file is part of Network-inador
 *
 * Copyright (C) 2022 - Félix Arreola Rodríguez
 *
 * Network-inador is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Network-inador is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Network-inador; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <stdint.h>

#include <linux/rtnetlink.h>

#include <netlink/socket.h>
#include <netlink/msg.h>

#include "common.h"
#include "routes.h"
#include "manager.h"

Route *_route_search_route (GList *list_routes, sa_family_t family, uint8_t type, uint32_t table, void *dest, uint32_t prefix, uint32_t priority) {
	GList *g;
	Route *route;
	int family_size = 0;
	
	if (family == AF_INET) {
		family_size = sizeof (struct in_addr);
	} else if (family == AF_INET6) {
		family_size = sizeof (struct in6_addr);
	}
	
	for (g = list_routes; g != NULL; g = g->next) {
		route = (Route *) g->data;
		
		if (route->family != family) continue;
		if (route->type != type) continue;
		if (route->table != table) continue;
		if (route->priority != priority) continue;
		
		if (memcmp (&route->dest, dest, family_size) == 0 && route->prefix == prefix) {
			return route;
		}
	}
	
	return NULL;
}

int routes_receive_message_newroute (struct nl_msg *msg, void *arg) {
	NetworkInadorHandle *handle = (NetworkInadorHandle *) arg;
	struct nlmsghdr *reply;
	Route *route = NULL;
	uint32_t table, prefix;
	int remaining, remaining2;
	int family, family_size = 0;
	struct rtmsg *rtm_hdr;
	struct nlattr *attr, *nest_attr_gw;
	RouteNH *next_hop;
	struct rtnexthop *nhptr;
	int rtnhp_len;
	int multipath_len;
	unsigned char *p;
	uint8_t route_type;
	struct_addr dest;
	uint32_t priority = 0;
	int was_new = 0;
	GList *route_list = NULL;
	
	reply = nlmsg_hdr (msg);
	
	if (reply->nlmsg_type != RTM_NEWROUTE) return NL_SKIP;
	
	/* Recuperar la tabla */
	rtm_hdr = nlmsg_data (reply);
	table = rtm_hdr->rtm_table;
	
	/* Recuperar la familia */
	family = rtm_hdr->rtm_family;
	if (family == AF_INET) {
		family_size = sizeof (struct in_addr);
		route_list = handle->route_v4_tables;
	} else if (family == AF_INET6) {
		family_size = sizeof (struct in6_addr);
		route_list = handle->route_v6_tables;
	}
	
	/* Recuperar el tipo de ruta */
	route_type = rtm_hdr->rtm_type;
	
	/* Recuperar el prefijo */
	prefix = rtm_hdr->rtm_dst_len;
	
	/* Recuperar solo el destino */
	memset (&dest, 0, sizeof (dest));
	nlmsg_for_each_attr(attr, reply, sizeof (struct rtmsg), remaining) {
		switch (nla_type (attr)) {
			case RTA_DST:
				if (nla_len (attr) != family_size) {
					/* Tamaño incorrecto para la IP */
					continue;
				}
				memcpy (&dest, nla_data (attr), family_size);
				break;
			case RTA_PRIORITY:
				if (nla_len (attr) != 4) {
					/* Tamaño incorrecto para la prioridad */
					continue;
				}
				priority = nla_get_u32 (attr);
				break;
			case RTA_TABLE:
				if (nla_len (attr) != 4) {
					/* Tamaño incorrecto para el id de la tabla */
					continue;
				}
				table = nla_get_u32 (attr);
				break;
		}
	}
	/* TODO: Revisar si la ruta que procesamos ya existe */
	route = _route_search_route (route_list, family, route_type, table, &dest, prefix, priority);
	
	if (route == NULL) {
		/* Nueva ruta */
		route = (Route *) malloc (sizeof (Route));
		memset (route, 0, sizeof (Route));
		
		route->family = family;
		route->type = route_type;
		route->table = table;
		route->prefix = prefix;
		
		route->priority = priority;
		memcpy (&route->dest, &dest, family_size);
		
		if (family == AF_INET) {
			handle->route_v4_tables = g_list_append (handle->route_v4_tables, route);
		} else if (family == AF_INET6) {
			handle->route_v6_tables = g_list_append (handle->route_v6_tables, route);
		}
		was_new = 1;
	} else {
		/* Liberar los next-hops, puesto que volverán a ser creados */
		g_list_free_full (route->nexthops, free);
		route->nexthops = NULL;
	}
	
	route->protocol = rtm_hdr->rtm_protocol;
	route->tos = rtm_hdr->rtm_tos;
	route->scope = rtm_hdr->rtm_scope;
	
	/* Pre-reservar el primer siguiente brinco y ligar */
	next_hop = (RouteNH *) malloc (sizeof (RouteNH));
	memset (next_hop, 0, sizeof (RouteNH));
	
	nlmsg_for_each_attr(attr, reply, sizeof (struct rtmsg), remaining) {
		switch (nla_type (attr)) {
			case RTA_PREFSRC:
				if (nla_len (attr) != family_size) {
					continue;
				}
				memcpy (&route->prefsrc, nla_data (attr), family_size);
				break;
			case RTA_GATEWAY:
				if (nla_len (attr) != family_size) {
					continue;
				}
				memcpy (&next_hop->gw, nla_data (attr), family_size);
				break;
			case RTA_OIF:
				if (nla_len (attr) != 4) {
					/* Tamaño incorrecto para el índice de la interfaz */
					continue;
				}
				next_hop->out_index = nla_get_u32 (attr);
				break;
			/* TODO: Revisar si RTA_PREF, RTA_CACHEINFO es útil */
			case RTA_MULTIPATH:
				nhptr = (struct rtnexthop*) nla_data (attr);
				multipath_len = nla_len (attr);
				
				while (multipath_len > 0) {
					next_hop->nh_flags = nhptr->rtnh_flags;
					next_hop->nh_weight = nhptr->rtnh_hops;
					next_hop->out_index = nhptr->rtnh_ifindex;
					
					/* Revisar en los atributos si tiene gw */
					p = ((unsigned char *) nhptr) + sizeof (struct rtnexthop);
					nest_attr_gw = nla_find ((const struct nlattr *) p, nhptr->rtnh_len - sizeof (struct rtnexthop), RTA_GATEWAY);
					
					if (nest_attr_gw != NULL && nla_len (nest_attr_gw) == family_size) {
						memcpy (&next_hop->gw, nla_data (nest_attr_gw), family_size);
					}
					
					multipath_len -= nhptr->rtnh_len;
					nhptr = (struct rtnexthop *) (((unsigned char *) nhptr) + nhptr->rtnh_len);
					
					/* Reservar el siguiente next-hop */
					if (multipath_len > 0) {
						route->nexthops = g_list_append (route->nexthops, next_hop);
						
						next_hop = (RouteNH *) malloc (sizeof (RouteNH));
						memset (next_hop, 0, sizeof (RouteNH));
					}
				}
				break;
		}
	}
	
	route->nexthops = g_list_append (route->nexthops, next_hop);
	
	if (was_new) {
		/* Enviar aquí evento de ruta agregada */
		manager_send_event_route_add (handle, route);
	} else {
		/* Enviar actualización */
		manager_send_event_route_update (handle, route);
	}
	return NL_SKIP;
}

int routes_receive_message_delroute (struct nl_msg *msg, void *arg) {
	NetworkInadorHandle *handle = (NetworkInadorHandle *) arg;
	struct nlmsghdr *reply;
	Route *route = NULL;
	uint32_t table, prefix;
	int remaining;
	int family, family_size = 0;
	struct rtmsg *rtm_hdr;
	struct nlattr *attr;
	uint8_t route_type;
	struct_addr dest;
	uint32_t priority = 0;
	GList *route_list = NULL;
	
	reply = nlmsg_hdr (msg);
	if (reply->nlmsg_type != RTM_DELROUTE) return NL_SKIP;
	
	/* Recuperar la tabla */
	rtm_hdr = nlmsg_data (reply);
	table = rtm_hdr->rtm_table;
	
	/* Recuperar la familia */
	family = rtm_hdr->rtm_family;
	if (family == AF_INET) {
		family_size = sizeof (struct in_addr);
		route_list = handle->route_v4_tables;
	} else if (family == AF_INET6) {
		family_size = sizeof (struct in6_addr);
		route_list = handle->route_v6_tables;
	}
	
	/* Recuperar el tipo de ruta */
	route_type = rtm_hdr->rtm_type;
	
	/* Recuperar el prefijo */
	prefix = rtm_hdr->rtm_dst_len;
	
	/* Recuperar solo el destino */
	memset (&dest, 0, sizeof (dest));
	nlmsg_for_each_attr(attr, reply, sizeof (struct rtmsg), remaining) {
		switch (nla_type (attr)) {
			case RTA_DST:
				if (nla_len (attr) != family_size) {
					/* Tamaño incorrecto para la IP */
					continue;
				}
				memcpy (&dest, nla_data (attr), family_size);
				break;
			case RTA_PRIORITY:
				if (nla_len (attr) != 4) {
					/* Tamaño incorrecto para la prioridad */
					continue;
				}
				priority = nla_get_u32 (attr);
				break;
			case RTA_TABLE:
				if (nla_len (attr) != 4) {
					/* Tamaño incorrecto para el id de la tabla */
					continue;
				}
				table = nla_get_u32 (attr);
				break;
		}
	}
	
	route = _route_search_route (route_list, family, route_type, table, &dest, prefix, priority);
	
	if (route == NULL) {
		/* ¿Notificación de eliminar una ruta que no existe? Super raro */
		return NL_SKIP;
	}
	
	if (family == AF_INET) {
		/* Eliminar de la lista ligada */
		handle->route_v4_tables = g_list_remove (handle->route_v4_tables, route);
	} else if (family == AF_INET6) {
		handle->route_v6_tables = g_list_remove (handle->route_v6_tables, route);
	}
	
	/* Notificar del evento */
	manager_send_event_route_del (handle, route);
	
	/* Eliminar todos los next-hops primero */
	g_list_free_full (route->nexthops, free);
	
	free (route);
	
	return NL_SKIP;
}

void routes_init (NetworkInadorHandle *handle) {
	/* Si es la primera vez que nos llaman, descargar una primera lista de interfaces */
	struct nl_msg * msg;
	struct rtmsg rtm_hdr = {
		.rtm_family = AF_UNSPEC,
	};
	int ret;
	
	msg = nlmsg_alloc_simple (RTM_GETROUTE, NLM_F_REQUEST | NLM_F_DUMP);
	if (msg == NULL) {
		return;
	}
	ret = nlmsg_append (msg, &rtm_hdr, sizeof (rtm_hdr), NLMSG_ALIGNTO);
	
	if (ret != 0) {
		nlmsg_free (msg);
		return;
	}
	
	nl_complete_msg (handle->nl_sock_route, msg);
	
	ret = nl_send (handle->nl_sock_route, msg);
	
	nlmsg_free (msg);
	
	nl_socket_modify_cb (handle->nl_sock_route, NL_CB_VALID, NL_CB_CUSTOM, routes_receive_message_newroute, handle);
	
	nl_recvmsgs_default (handle->nl_sock_route);
}

void routes_clean_up (NetworkInadorHandle *handle) {
	GList *g;
	Route *route;
	
	for (g = handle->route_v4_tables; g != NULL; g = g->next) {
		route = (Route *) g->data;
		
		g_list_free_full (route->nexthops, free);
		route->nexthops = NULL;
	}
	
	g_list_free_full (handle->route_v4_tables, free);
	handle->route_v4_tables = NULL;
	
	for (g = handle->route_v6_tables; g != NULL; g = g->next) {
		route = (Route *) g->data;
		
		g_list_free_full (route->nexthops, free);
		route->nexthops = NULL;
	}
	
	g_list_free_full (handle->route_v6_tables, free);
	handle->route_v6_tables = NULL;
}

